
import java.util.*;
import java.text.*;

class SimpleMVA 
{
   public static void main (String args[]) {

/*
              ***** First part defines the queueing network  *****
          ***** This is the only part that is should be altered   *****
*/

//            Example from Notes  ============>
      int M = 5;                                   // Number of nodes
      int N = 10;                                  // Population size
      int [] qType = {0,1,2,2,2,2};                // Node types (1, 2 or 3)
      int [] mnum = {0,10,1,1,1,2};                // Number of servers per node 
      double [] mu = {0,2,1,2,4,4.5};              // Service rates for each node
      double [] V  = {0,1.0,0.2,0.3,0.5,1.0};      // Visit ratios for each node
/*
//            Example from Bolch et al.  ============>
      int M = 4;
      int N = 3;
      int [] qType = {0,3,2,2,1};
      int [] mnum =  {0,2,1,1,3};
      double [] mu = { 0, 2, 1.66666666666, 1.25, 1};
      double [] V =  {0, 1.0, 0.5, 0.5, 1.0}; 
*/
/*
                      ***** The MVA algorithm  *****
          ***** There is no need to alter beyond this point  ****
*/
     
      double G = 1;                                // Normalization constant
      double [] L = new double [M+1];              // Mean number of customers
      double [] R = new double [M+1];              // Mean response times
      double [] X = new double [M+1];              // Throughputs

      int numLD = 0;                    
      for (int i=1; i<=M; i++) {
          if (qType[i] == 3) {numLD++;}            // Determine number of LD nodes
      }
      double [][] pp = new double [N+1][numLD+1];
      for (int m=0; m<=numLD; m++) {
         pp[0][m] = 1;                             // Marginals for LD nodes
      }




//               ***** Begin iterating over customer population  ***
      for (int k=1; k<=N; k++){

//       ***  MVA Step 1:  Compute mean response times  ***
         int LDcount = 0;                                   // Indicator for LD nodes
         for (int m=1; m <= M; m++) {
            if (qType[m] == 1) { R[m] = 1/mu[m]; }          // Infinite server node
            if (qType[m] == 2) { R[m] = (L[m]+1)/mu[m]; }   // FCFS exponential, etc.
            if (qType[m] == 3) {                            // Load-dependent node
               R[m] = 0;                                    //   ... considered to be a
               for (int j=1; j <=k; j++) {                  //   multiserver expon. node
                   // In this implementation, alpha_i(j) is equal to min(m_i, j)
                   double alpha = j; if (mnum[m] < j){alpha = mnum[m];}
                   R[m] = R[m] + j/(mu[m] * alpha) * pp[j-1][LDcount];
               }
               LDcount++;
            }
         }     

//       ***  MVA Step 2:   Compute throughputs
         double bottom = 0;
         for (int m=1; m<=M; m++){
             bottom = bottom + V[m]*R[m];
         }
         X[1] = k/bottom;                          // Overall throughput
         for (int m=2; m<=M; m++) {
             X[m] = V[m]*X[1];                     // Throughput of node m
         }
//                  Update normalization constant
         G = G/X[1];

//       *** MVA Step 3:   Compute mean number of customers  ***
         for (int m=1; m<= M; m++) {
            L[m] = X[m]*R[m];
         }

//               Compute marginals for all load-dependent nodes
         LDcount = 0;                              //  Indicator for LD nodes
         for (int m=1; m<=M; m++) {
             if (qType[m] == 3) {
                double p0 = 0;
                for (int j=k; j >=1; j--) {
                    // In this implementation, alpha_i(j) is equal to min(m_i, j)
                    double alpha = j; if (mnum[m] < j){alpha = mnum[m];}
                    pp[j][LDcount] = X[1]/(mu[m] * alpha) * pp[j-1][LDcount];
                    p0 = p0 + pp[j][LDcount];
                } 
                pp[0][LDcount] = 1-p0;
                LDcount++;
            }
         }
      }
   
//              Output results
      NumberFormat fmat = NumberFormat.getNumberInstance();
      fmat.setMaximumFractionDigits(6);
      fmat.setMinimumFractionDigits(6);
      System.out.println("     " );
      System.out.println("              R[m]       X[m]        L[m] " );
      for (int m=1; m <=M; m++) {
         System.out.println(" m = " + m + ":     " + fmat.format(R[m]) + "   " 
                      + fmat.format(X[m]) + "    " + fmat.format(L[m]));
      } 
      System.out.println(" " );
      double sum = 0;
      for (int m=1; m <=M; m++) {
          sum = sum + L[m];
      }
      System.out.println("Sum check:  Population size =  "  + fmat.format(sum));
      System.out.println("Normalization constant: G = " + G);
      System.out.println("     " );
   }
}


